#include "X3D_PrimitiveMeshBuilderGeometry3DVisitor.h"
#include "X3D_Box.h"
#include "X3D_Cylinder.h"
#include "X3D_Cone.h"
#include "X3D_Sphere.h"
#include "MESH_SceneGraphTypes.h"
#include "MESH_Shape.h"
#include "MESH_Mesh.h"
#include "MESH_Vertex.h"
#include "MESH_PrimitiveMeshBuilderStateVariables.h"

#include <vector>

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

PrimitiveMeshBuilderGeometry3DVisitor::PrimitiveMeshBuilderGeometry3DVisitor()
: Geometry3DVisitor()
{
  define(Recorder<Box>::getEnterFunction(&PrimitiveMeshBuilderGeometry3DVisitor::enterBox));
  define(Recorder<Cylinder>::getEnterFunction(&PrimitiveMeshBuilderGeometry3DVisitor::enterCylinder));
  define(Recorder<Cone>::getEnterFunction(&PrimitiveMeshBuilderGeometry3DVisitor::enterCone));
  define(Recorder<Sphere>::getEnterFunction(&PrimitiveMeshBuilderGeometry3DVisitor::enterSphere));
}

void PrimitiveMeshBuilderGeometry3DVisitor::enterBox(Box *B)
{
  MESH::PrimitiveMeshBuilderStateVariables *stateVariables = Singleton<MESH::PrimitiveMeshBuilderStateVariables>::getInstance();
  if (!stateVariables->getNode(B))
  {
    // Creating a new MESH node.
    MESH::Vertex *V = new MESH::Vertex();
    MESH::Mesh *M = new MESH::Mesh();
    M->setChild(V);
    
    M->data().setColor(false);
    M->data().setNormal(true);
    M->data().setCreaseAngle(1.05f);
    M->data().setNormalPerVertex(true);
    M->data().setSolid(true);
    M->data().setTexCoord(false);
    
    // filling the nodes
    
    MESH::SFVertex *v0 = V->createVertex();
    MESH::SFVertex *v1 = V->createVertex();
    MESH::SFVertex *v2 = V->createVertex();
    MESH::SFVertex *v3 = V->createVertex();
    MESH::SFVertex *v4 = V->createVertex();
    MESH::SFVertex *v5 = V->createVertex();
    MESH::SFVertex *v6 = V->createVertex();
    MESH::SFVertex *v7 = V->createVertex();
    
    SFVec3f size = 0.5f*B->getSize();
    
    v0->data().setPoint(SFPoint3f(-size.x, -size.y,  size.z));
    v1->data().setPoint(SFPoint3f( size.x, -size.y,  size.z));
    v2->data().setPoint(SFPoint3f( size.x,  size.y,  size.z));
    v3->data().setPoint(SFPoint3f(-size.x,  size.y,  size.z));

    v4->data().setPoint(SFPoint3f(-size.x, -size.y, -size.z));
    v5->data().setPoint(SFPoint3f( size.x, -size.y, -size.z));
    v6->data().setPoint(SFPoint3f( size.x,  size.y, -size.z));
    v7->data().setPoint(SFPoint3f(-size.x,  size.y, -size.z));
    
    MFInt32 front, back, left, right, top, bottom;
    
    front.push_back(v0->getIndex());
    front.push_back(v1->getIndex());
    front.push_back(v2->getIndex());
    front.push_back(v3->getIndex());
    
    back.push_back(v5->getIndex());
    back.push_back(v4->getIndex());
    back.push_back(v7->getIndex());
    back.push_back(v6->getIndex());
    
    left.push_back(v4->getIndex());
    left.push_back(v0->getIndex());
    left.push_back(v3->getIndex());
    left.push_back(v7->getIndex());
    
    right.push_back(v1->getIndex());
    right.push_back(v5->getIndex());
    right.push_back(v6->getIndex());
    right.push_back(v2->getIndex());
    
    top.push_back(v3->getIndex());
    top.push_back(v2->getIndex());
    top.push_back(v6->getIndex());
    top.push_back(v7->getIndex());
    
    bottom.push_back(v4->getIndex());
    bottom.push_back(v5->getIndex());
    bottom.push_back(v1->getIndex());
    bottom.push_back(v0->getIndex());
    
    M->createFace(front);
    M->createFace(back);
    M->createFace(left);
    M->createFace(right);
    M->createFace(top);
    M->createFace(bottom);
    
    // Managing scene graphs links
    
    MFNode parentList = B->getParentList();
    for (MFNode::const_iterator p = parentList.begin(); p != parentList.end(); ++p)
    {
      MESH::Shape *S = new MESH::Shape();
      S->setChild(M);
      
      MFNode pparentList = (*p)->getParentList();
      for (MFNode::const_iterator pp = pparentList.begin(); pp != pparentList.end(); ++pp)
      {
        (*pp)->removeChild(*p);
        (*pp)->setChild(S);
      }  
      
      stateVariables->addToDeleteNode(*p);
    }  
    
    stateVariables->addNode(B);
  }
}

void PrimitiveMeshBuilderGeometry3DVisitor::enterCylinder(Cylinder *C)
{
  MESH::PrimitiveMeshBuilderStateVariables *stateVariables = Singleton<MESH::PrimitiveMeshBuilderStateVariables>::getInstance();
  if (!stateVariables->getNode(C))
  {
    // Creating a new MESH node.
    MESH::Vertex *V = new MESH::Vertex();
    MESH::Mesh *M = new MESH::Mesh();
    M->setChild(V);
    
    M->data().setColor(false);
    M->data().setNormal(true);
    M->data().setCreaseAngle(1.05f);
    M->data().setNormalPerVertex(true);
    M->data().setSolid(true);
    M->data().setTexCoord(false);
    
    unsigned int section = static_cast<unsigned int>(stateVariables->getStep());
    float cf = -2.0f*(float)PI/(float)section;
    
    SFFloat height = 0.5f*C->getHeight();
    SFFloat radius = C->getRadius();
    
    // filling the nodes
    vector<MESH::SFVertex *> vertexBottomArray;
    vector<MESH::SFVertex *> vertexTopArray;
    
    MESH::SFVertex *b = V->createVertex();
    b->data().setPoint(SFPoint3f(0.0f, -height, 0.0f));
    MESH::SFVertex *t = V->createVertex();
    t->data().setPoint(SFPoint3f(0.0f, height, 0.0f));
    
    for (unsigned int i = 0; i < section; ++i)
    {
      MESH::SFVertex *vb = V->createVertex();
      vb->data().setPoint(SFPoint3f(radius*cosf(cf*(float)i), -height, radius*sinf(cf*(float)i)));
      
      MESH::SFVertex *vt = V->createVertex();
      vt->data().setPoint(SFPoint3f(radius*cosf(cf*(float)i), height, radius*sinf(cf*(float)i)));
      
      vertexBottomArray.push_back(vb);
      vertexTopArray.push_back(vt);
    }
    
    for (unsigned int j = 0; j < section; ++j)
    {
      if (C->getBottom())
      {
        MFInt32 bottom;
        bottom.push_back(b->getIndex());
        bottom.push_back(vertexBottomArray[(j + 1) % section]->getIndex());
        bottom.push_back(vertexBottomArray[j]->getIndex());
        M->createFace(bottom);
      }  
      
      if (C->getSide())
      {
        MFInt32 side0, side1;
        side0.push_back(vertexBottomArray[j]->getIndex());
        side0.push_back(vertexBottomArray[(j + 1) % section]->getIndex());
        side0.push_back(vertexTopArray[j]->getIndex());
        M->createFace(side0);
        
      
        side1.push_back(vertexBottomArray[(j + 1) % section]->getIndex());
        side1.push_back(vertexTopArray[(j + 1) % section]->getIndex());
        side1.push_back(vertexTopArray[j]->getIndex());
        M->createFace(side1);
      }  
      
      if (C->getTop())
      {
        MFInt32 top;
        top.push_back(t->getIndex());
        top.push_back(vertexTopArray[j]->getIndex());
        top.push_back(vertexTopArray[(j + 1) % section]->getIndex());
        M->createFace(top);
      }  
    }
        
    // Managing scene graphs links
    
    MFNode parentList = C->getParentList();
    for (MFNode::const_iterator p = parentList.begin(); p != parentList.end(); ++p)
    {
      MESH::Shape *S = new MESH::Shape();
      S->setChild(M);
      
      MFNode pparentList = (*p)->getParentList();
      for (MFNode::const_iterator pp = pparentList.begin(); pp != pparentList.end(); ++pp)
      {
        (*pp)->removeChild(*p);
        (*pp)->setChild(S);
      }  
      
      stateVariables->addToDeleteNode(*p);
    }  
    
    stateVariables->addNode(C);
  }
}

void PrimitiveMeshBuilderGeometry3DVisitor::enterCone(Cone *C)
{
  MESH::PrimitiveMeshBuilderStateVariables *stateVariables = Singleton<MESH::PrimitiveMeshBuilderStateVariables>::getInstance();
  if (!stateVariables->getNode(C))
  {
    // Creating a new MESH node.
    MESH::Vertex *V = new MESH::Vertex();
    MESH::Mesh *M = new MESH::Mesh();
    M->setChild(V);
    
    M->data().setColor(false);
    M->data().setNormal(true);
    M->data().setCreaseAngle(1.05f);
    M->data().setNormalPerVertex(true);
    M->data().setSolid(true);
    M->data().setTexCoord(false);
    
    unsigned int section = static_cast<unsigned int>(stateVariables->getStep());
    float cf = -2.0f*(float)PI/(float)section;
    SFFloat radius = C->getBottomRadius();
    SFFloat height = 0.5f*C->getHeight();
    
    // filling the nodes
    vector<MESH::SFVertex *> vertexBottomArray;
    
    MESH::SFVertex *b = V->createVertex();
    b->data().setPoint(SFPoint3f(0.0f, -height, 0.0f));
    MESH::SFVertex *t = V->createVertex();
    t->data().setPoint(SFPoint3f(0.0f, height, 0.0f));
    
    for (unsigned int i = 0; i < section; ++i)
    {
      MESH::SFVertex *vb = V->createVertex();
      vb->data().setPoint(SFPoint3f(radius*cosf(cf*(float)i), -height, radius*sinf(cf*(float)i)));
      
      vertexBottomArray.push_back(vb);
    }
    
    for (unsigned int j = 0; j < section; ++j)
    {
      if (C->getBottom())
      {
        MFInt32 bottom;
        bottom.push_back(b->getIndex());
        bottom.push_back(vertexBottomArray[(j + 1) % section]->getIndex());
        bottom.push_back(vertexBottomArray[j]->getIndex());
        M->createFace(bottom);
      }  
      
      if (C->getSide())
      {
        MFInt32 side;
        side.push_back(vertexBottomArray[j]->getIndex());
        side.push_back(vertexBottomArray[(j + 1) % section]->getIndex());
        side.push_back(t->getIndex());
        M->createFace(side);
      }  
    }
        
    // Managing scene graphs links
    
    MFNode parentList = C->getParentList();
    for (MFNode::const_iterator p = parentList.begin(); p != parentList.end(); ++p)
    {
      MESH::Shape *S = new MESH::Shape();
      S->setChild(M);
      
      MFNode pparentList = (*p)->getParentList();
      for (MFNode::const_iterator pp = pparentList.begin(); pp != pparentList.end(); ++pp)
      {
        (*pp)->removeChild(*p);
        (*pp)->setChild(S);
      }  
      
      stateVariables->addToDeleteNode(*p);
    }  
    
    stateVariables->addNode(C);
  }
}

void PrimitiveMeshBuilderGeometry3DVisitor::enterSphere(Sphere *S)
{
  MESH::PrimitiveMeshBuilderStateVariables *stateVariables = Singleton<MESH::PrimitiveMeshBuilderStateVariables>::getInstance();
  if (!stateVariables->getNode(S))
  {
    // Creating a new MESH node.
    MESH::Vertex *V = new MESH::Vertex();
    MESH::Mesh *M = new MESH::Mesh();
    M->setChild(V);
    
    M->data().setColor(false);
    M->data().setNormal(true);
    M->data().setCreaseAngle(1.05f);
    M->data().setNormalPerVertex(true);
    M->data().setSolid(true);
    M->data().setTexCoord(false);
    
    unsigned int section = static_cast<unsigned int>(stateVariables->getStep());
    unsigned int msection = section/2;
    SFFloat radius = S->getRadius();
    
    // filling the nodes
    vector<MESH::SFVertex *> vertexArray;
    
    MESH::SFVertex *b = V->createVertex();
    b->data().setPoint(SFPoint3f(0.0f, -radius, 0.0f));
    MESH::SFVertex *t = V->createVertex();
    t->data().setPoint(SFPoint3f(0.0f, radius, 0.0f));
    
    for (unsigned int j = 0; j < section; ++j)
    {
      for (unsigned int i = 0; i < msection; ++i)
      {
        MESH::SFVertex *v = V->createVertex();
        
        float phi = (float)PI*(float)i/(float)msection;
        float theta = 2.0f*(float)PI*(float)j/(float)section;
        
        v->data().setPoint(radius*SFPoint3f(sinf(theta)*sinf(phi), cosf(phi), cosf(theta)*sinf(phi)));
        vertexArray.push_back(v);
      }  
    }
    
    for (unsigned int k = 0; k < section; ++k)
    {
      MFInt32 face0;
      face0.push_back(t->getIndex());
      face0.push_back(vertexArray[k*msection + 1]->getIndex());
      face0.push_back(vertexArray[((k + 1) % section)*msection + 1]->getIndex());
      M->createFace(face0);
    }    
    
    for (unsigned int l = 0; l < section; ++l)
    {
      for (unsigned int m = 1; m < msection - 1; ++m)
      {
        MFInt32 face0, face1;
        face0.push_back(vertexArray[l*msection + m]->getIndex());
        face0.push_back(vertexArray[l*msection + ((m + 1) % msection)]->getIndex());
        face0.push_back(vertexArray[((l + 1) % section)*msection + ((m + 1) % msection)]->getIndex());
        M->createFace(face0);
        
        face1.push_back(vertexArray[l*msection + m]->getIndex());
        face1.push_back(vertexArray[((l + 1) % section)*msection + ((m + 1) % msection)]->getIndex());
        face1.push_back(vertexArray[((l + 1) % section)*msection + m]->getIndex());
        M->createFace(face1);
      }
    }  
    
    for (unsigned int n = 0; n < section; ++n)
    {
      MFInt32 face0;
      face0.push_back(vertexArray[n*msection + msection - 1]->getIndex());
      face0.push_back(b->getIndex());
      face0.push_back(vertexArray[((n + 1) % section)*msection + msection - 1]->getIndex());
      M->createFace(face0);
    }
        
    // Managing scene graphs links
    
    MFNode parentList = S->getParentList();
    for (MFNode::const_iterator p = parentList.begin(); p != parentList.end(); ++p)
    {
      MESH::Shape *Sh = new MESH::Shape();
      Sh->setChild(M);
      
      MFNode pparentList = (*p)->getParentList();
      for (MFNode::const_iterator pp = pparentList.begin(); pp != pparentList.end(); ++pp)
      {
        (*pp)->removeChild(*p);
        (*pp)->setChild(Sh);
      }  
      
      stateVariables->addToDeleteNode(*p);
    }  
    
    stateVariables->addNode(S);
  }
}
  
}
}
